@@ -1,7 +1,7 @@
 %%
 %% %CopyrightBegin%
 %%
-%% Copyright Ericsson AB 2012-2016. All Rights Reserved.
+%% Copyright Ericsson AB 2012-2021. All Rights Reserved.
 %%
 %% Licensed under the Apache License, Version 2.0 (the "License");
 %% you may not use this file except in compliance with the License.
@@ -18,7 +18,7 @@
 %% %CopyrightEnd%
 -module(observer_perf_wx).
 
--export([start_link/2]).
+-export([start_link/3]).
 
 %% wx_object callbacks
 -export([init/1, handle_info/2, terminate/2, code_change/3, handle_call/3,
@@ -55,12 +55,12 @@
 
 -define(wxGC, wxGraphicsContext).
 
--record(paint, {font, small, pen, pen2, pens, usegc = false}).
+-record(paint, {font, small, fg, pen, pen2, pens, dot_pens, usegc = false}).
 
-start_link(Notebook, Parent) ->
-    wx_object:start_link(?MODULE, [Notebook, Parent], []).
+start_link(Notebook, Parent, Config) ->
+    wx_object:start_link(?MODULE, [Notebook, Parent, Config], []).
 
-init([Notebook, Parent]) ->
+init([Notebook, Parent, Config]) ->
     try
 	Panel = wxPanel:new(Notebook),
 	Main  = wxBoxSizer:new(?wxVERTICAL),
@@ -81,11 +81,13 @@ init([Notebook, Parent]) ->
 			panel =Panel,
 			wins = Windows,
 			paint=PaintInfo,
-			samples=reset_data()
+			samples=reset_data(),
+                        time=#ti{fetch=maps:get(fetch, Config, ?FETCH_DATA),
+                                 secs=maps:get(secs, Config, ?DISP_SECONDS)}
 		       },
 	{Panel, State0}
-    catch _:Err ->
-	    io:format("~p crashed ~p: ~p~n",[?MODULE, Err, erlang:get_stacktrace()]),
+    catch _:Err:Stacktrace ->
+	    io:format("~p crashed ~tp: ~tp~n",[?MODULE, Err, Stacktrace]),
 	    {stop, Err}
     end.
 
@@ -97,40 +99,51 @@ make_win(Name, Parent, Sizer, Border) ->
     #win{name=Name, panel=Panel}.
 
 setup_graph_drawing(Panels) ->
-    IsWindows = element(1, os:type()) =:= win32,
-    IgnoreCB = {callback, fun(_,_) -> ok end},
     Do = fun(#win{panel=Panel}) ->
-		 wxWindow:setBackgroundStyle(Panel, ?wxBG_STYLE_SYSTEM),
-		 wxPanel:connect(Panel, paint, [callback]),
-		 IsWindows andalso
-		     wxPanel:connect(Panel, erase_background, [IgnoreCB])
+		 wxWindow:setBackgroundStyle(Panel, ?wxBG_STYLE_PAINT),
+		 wxPanel:connect(Panel, paint, [callback])
 	 end,
     _ = [Do(Panel) || Panel <- Panels],
     UseGC = haveGC(),
     Version28 = ?wxMAJOR_VERSION =:= 2 andalso ?wxMINOR_VERSION =:= 8,
+    Scale = observer_wx:get_scale(),
     {Font, SmallFont}
 	= if UseGC, Version28 ->
 		  %% Def font is really small when using Graphics contexts in 2.8
 		  %% Hardcode it
-		  F = wxFont:new(12,?wxFONTFAMILY_DECORATIVE,?wxFONTSTYLE_NORMAL,?wxFONTWEIGHT_BOLD),
-		  SF = wxFont:new(10, ?wxFONTFAMILY_DECORATIVE, ?wxFONTSTYLE_NORMAL, ?wxFONTWEIGHT_NORMAL),
+		  F = wxFont:new(Scale * 12,?wxFONTFAMILY_DECORATIVE,?wxFONTSTYLE_NORMAL,?wxFONTWEIGHT_BOLD),
+		  SF = wxFont:new(Scale * 10, ?wxFONTFAMILY_DECORATIVE, ?wxFONTSTYLE_NORMAL, ?wxFONTWEIGHT_NORMAL),
 		  {F, SF};
 	     true ->
 		  DefFont = wxSystemSettings:getFont(?wxSYS_DEFAULT_GUI_FONT),
 		  DefSize = wxFont:getPointSize(DefFont),
 		  DefFamily = wxFont:getFamily(DefFont),
-		  F = wxFont:new(DefSize-1, DefFamily, ?wxFONTSTYLE_NORMAL, ?wxFONTWEIGHT_BOLD),
-		  SF = wxFont:new(DefSize-2, DefFamily, ?wxFONTSTYLE_NORMAL, ?wxFONTWEIGHT_NORMAL),
+		  F = wxFont:new(Scale * (DefSize-1), DefFamily, ?wxFONTSTYLE_NORMAL, ?wxFONTWEIGHT_BOLD),
+		  SF = wxFont:new(Scale * (DefSize-2), DefFamily, ?wxFONTSTYLE_NORMAL, ?wxFONTWEIGHT_NORMAL),
 		  {F, SF}
 	  end,
-    BlackPen = wxPen:new({0,0,0}, [{width, 1}]),
-    Pens = [wxPen:new(Col, [{width, 1}]) || Col <- tuple_to_list(colors())],
+    BG = wxWindow:getBackgroundColour((hd(Panels))#win.panel),
+    Fg = case observer_lib:is_darkmode(BG) of
+             false -> {0,0,0};
+             true  -> wxSystemSettings:getColour(?wxSYS_COLOUR_BTNTEXT)
+         end,
+
+    PenColor = case observer_lib:is_darkmode(BG) of
+                   false -> {0,0,0};
+                   true  -> {0,0,0}
+               end,
+    Pens = [wxPen:new(Col, [{width, Scale}, {style, ?wxSOLID}])
+            || Col <- tuple_to_list(colors())],
+    DotPens = [wxPen:new(Col, [{width, Scale}, {style, ?wxDOT}])
+               || Col <- tuple_to_list(colors())],
     #paint{usegc = UseGC,
 	   font  = Font,
 	   small = SmallFont,
-	   pen   = ?wxGREY_PEN,
-	   pen2  = BlackPen,
-	   pens  = list_to_tuple(Pens)
+           fg    = Fg,  %% Text color
+	   pen   = wxPen:new(PenColor),
+	   pen2  = wxPen:new(PenColor, [{width, Scale}]),
+	   pens  = list_to_tuple(Pens),
+           dot_pens = list_to_tuple(DotPens)
 	  }.
 
 
@@ -173,6 +186,10 @@ refresh_panel(Active, #win{name=_Id, panel=Panel}=Win, Ti, #paint{usegc=UseGC}=P
     destroy_gc(GC).
 
 %%%%%%%%%%
+handle_call(get_config, _, #state{time=Ti}=State) ->
+    #ti{fetch=Fetch, secs=Range} = Ti,
+    {reply, #{fetch=>Fetch, secs=>Range}, State};
+
 handle_call(Event, From, _State) ->
     error({unhandled_call, Event, From}).
 
@@ -181,17 +198,17 @@ handle_cast(Event, _State) ->
 %%%%%%%%%%
 handle_info({stats, 1, _, _, _} = Stats,
 	    #state{panel=Panel, samples=Data, active=Active, wins=Wins0,
-		   time=#ti{tick=Tick, disp=Disp0}=Ti} = State0) ->
+                   appmon=Node, time=#ti{tick=Tick, disp=Disp0}=Ti} = State0) ->
     if Active ->
 	    Disp = trunc(Disp0),
 	    Next = max(Tick - Disp, 0),
 	    erlang:send_after(1000 div ?DISP_FREQ, self(), {refresh, Next}),
-	    {Wins, Samples} = add_data(Stats, Data, Wins0, Ti, Active),
+	    {Wins, Samples} = add_data(Stats, Data, Wins0, Ti, Active, Node),
 	    State = precalc(State0#state{time=Ti#ti{tick=Next}, wins=Wins, samples=Samples}),
 	    wxWindow:refresh(Panel),
 	    {noreply, State};
        true ->
-	    {Wins1, Samples} = add_data(Stats, Data, Wins0, Ti, Active),
+	    {Wins1, Samples} = add_data(Stats, Data, Wins0, Ti, Active, Node),
 	    Wins = [W#win{max=undefined} || W <- Wins1],
 	    {noreply, State0#state{samples=Samples, wins=Wins, time=Ti#ti{tick=0}}}
     end;
@@ -206,7 +223,7 @@ handle_info({refresh, Seq}, #state{panel=Panel, time=#ti{tick=Seq, disp=DispF}=T
 handle_info({refresh, _}, State) ->
     {noreply, State};
 
-handle_info({active, Node}, #state{parent=Parent, panel=Panel, appmon=Old, time=_Ti} = State) ->
+handle_info({active, Node}, #state{parent=Parent, panel=Panel, appmon=Old} = State) ->
     create_menus(Parent, []),
     try
 	Node = node(Old),
@@ -225,7 +242,7 @@ handle_info({'EXIT', Old, _}, State = #state{appmon=Old}) ->
     {noreply, State#state{active=false, appmon=undefined}};
 
 handle_info(_Event, State) ->
-    %% io:format("~p:~p: ~p~n",[?MODULE,?LINE,_Event]),
+    %% io:format("~p:~p: ~tp~n",[?MODULE,?LINE,_Event]),
     {noreply, State}.
 
 %%%%%%%%%%
@@ -235,39 +252,50 @@ terminate(_Event, #state{appmon=Pid}) ->
 code_change(_, _, State) ->
     State.
 
-restart_fetcher(Node, #state{appmon=Old, panel=Panel, time=#ti{fetch=Freq}=Ti}=State) ->
+restart_fetcher(Node, #state{appmon=Old, panel=Panel, time=#ti{fetch=Freq}=Ti, wins=Wins0}=State) ->
     catch Old ! exit,
     Me = self(),
     Pid = spawn_link(Node, observer_backend, fetch_stats, [Me, round(1000/Freq)]),
     wxWindow:refresh(Panel),
-    precalc(State#state{active=true, appmon=Pid, samples=reset_data(), time=Ti#ti{tick=0}}).
+    Wins = [W#win{state=undefined} || W <- Wins0],
+    precalc(State#state{active=true, appmon=Pid, samples=reset_data(),
+			wins=Wins, time=Ti#ti{tick=0}}).
 
 reset_data() ->
     {0, queue:new()}.
 
-add_data(Stats, {N, Q0}, Wins, #ti{fetch=Fetch, secs=Secs}, Active) when N > (Secs*Fetch+1) ->
-    {{value, Drop}, Q} = queue:out(Q0),
-    add_data_1(Wins, Stats, N, {Drop,Q}, Active);
-add_data(Stats, {N, Q}, Wins, _, Active) ->
-    add_data_1(Wins, Stats, N+1, {empty, Q}, Active).
+add_data(Stats, Q, Wins, Ti, Active) ->
+    add_data(Stats, Q, Wins, Ti, Active, ignore).
 
-add_data_1([#win{state={_,St}}|_]=Wins0, Last, N, {Drop, Q}, Active)
+add_data(Stats, {N, Q0}, Wins, #ti{fetch=Fetch, secs=Secs}, Active, Node)
+  when N > (Secs*Fetch+1) ->
+    {{value, Drop}, Q} = queue:out(Q0),
+    add_data_1(Wins, Stats, N, {Drop,Q}, Active, Node);
+add_data(Stats, {N, Q}, Wins, _, Active, Node) ->
+    add_data_1(Wins, Stats, N+1, {empty, Q}, Active, Node).
+
+add_data_1([#win{state={_,St}}|_]=Wins0, Last, N, {Drop, Q}, Active, Node)
   when St /= undefined ->
-    {Wins, Stat} =
-	lists:mapfoldl(fun(Win0, Entry) ->
-			       {Win1,Stat} = add_data_2(Win0, Last, Entry),
-			       case Active of
-				   true ->
-				       Win = add_data_3(Win1, N, Drop, Stat, Q),
-				       {Win, Stat};
-				   false ->
-				       {Win1, Stat}
-			       end
-		       end, #{}, Wins0),
-    {Wins, {N,queue:in(Stat#{}, Q)}};
-add_data_1(Wins, Stats, 1, {_, Q}, _) ->
-    {[Win#win{state=init_data(Id, Stats),
-	      info = info(Id, Stats)}
+    try
+	{Wins, Stat} =
+	    lists:mapfoldl(fun(Win0, Entry) ->
+				   {Win1,Stat} = add_data_2(Win0, Last, Entry),
+				   case Active of
+				       true ->
+					   Win = add_data_3(Win1, N, Drop, Stat, Q),
+					   {Win, Stat};
+				       false ->
+					   {Win1, Stat}
+				   end
+			   end, #{}, Wins0),
+	{Wins, {N,queue:in(Stat#{}, Q)}}
+    catch no_scheduler_change ->
+	    {[Win#win{state=init_data(Id, Last), info=info(Id, Last, Node)}
+	      || #win{name=Id}=Win <- Wins0], {0,queue:new()}}
+    end;
+
+add_data_1(Wins, Stats, 1, {_, Q}, _, Node) ->
+    {[Win#win{state=init_data(Id, Stats), info=info(Id, Stats, Node)}
       || #win{name=Id}=Win <- Wins], {0,Q}}.
 
 add_data_2(#win{name=Id, state=S0}=Win, Stats, Map) ->
@@ -373,16 +401,24 @@ lmax(MState, Values, State) ->
 
 init_data(runq, {stats, _, T0, _, _}) -> {mk_max(),lists:sort(T0)};
 init_data(io,   {stats, _, _, {{_,In0}, {_,Out0}}, _}) -> {mk_max(), {In0,Out0}};
-init_data(memory, _) -> {mk_max(), info(memory, undefined)};
-init_data(alloc, _) -> {mk_max(), ok};
-init_data(utilz, _) -> {mk_max(), ok}.
+init_data(memory, _) -> {mk_max(), info(memory, undefined, undefined)};
+init_data(alloc, _) -> {mk_max(), unused};
+init_data(utilz, _) -> {mk_max(), unused}.
 
-info(runq, {stats, _, T0, _, _}) -> lists:seq(1, length(T0));
-info(memory, _) -> [total, processes, atom, binary, code, ets];
-info(io, _) -> [input, output];
-info(alloc, First) -> [Type || {Type, _, _} <- First];
-info(utilz, First) -> [Type || {Type, _, _} <- First];
-info(_, []) -> [].
+info(runq, {stats, _, T0, _, _}, Node) ->
+    Dirty = get_dirty_cpu(Node),
+    {lists:seq(1, length(T0)-Dirty), Dirty};
+info(memory, _, _) -> [total, processes, atom, binary, code, ets];
+info(io, _, _) -> [input, output];
+info(alloc, First, _) -> [Type || {Type, _, _} <- First];
+info(utilz, First, _) -> [Type || {Type, _, _} <- First];
+info(_, [], _) -> [].
+
+get_dirty_cpu(Node) ->
+    case rpc:call(node(Node), erlang, system_info, [dirty_cpu_schedulers]) of
+        {badrpc,_R} -> 0;
+        N -> N
+    end.
 
 collect_data(runq, {stats, _, T0, _, _}, {Max,S0}) ->
     S1 = lists:sort(T0),
@@ -401,15 +437,16 @@ collect_data(memory, {stats, _, _, _, MemInfo}, {Max, MemTypes}) ->
 collect_data(alloc, MemInfo, Max) ->
     Vs = [Carrier || {_Type,_Block,Carrier} <- MemInfo],
     Sample = list_to_tuple(Vs),
-    {Sample, lmax(Max, Vs, Sample)};
+    {Sample, {lmax(Max, Vs, Sample),unused}};
 collect_data(utilz, MemInfo, Max) ->
     Vs = [round(100*Block/Carrier) || {_Type,Block,Carrier} <- MemInfo],
     Sample = list_to_tuple(Vs),
-    {Sample, lmax(Max,Vs,Sample)}.
+    {Sample, {lmax(Max,Vs,Sample),unused}}.
 
 calc_delta([{Id, WN, TN}|Ss], [{Id, WP, TP}|Ps]) ->
     [100*(WN-WP) div (TN-TP)|calc_delta(Ss, Ps)];
-calc_delta([], []) -> [].
+calc_delta([], []) -> [];
+calc_delta(_, _) -> throw(no_scheduler_change).
 
 precalc(#state{samples=Data0, paint=Paint, time=Ti, wins=Wins0}=State) ->
     Wins = [precalc(Ti, Data0, Paint, Win) || Win <- Wins0],
@@ -461,9 +498,10 @@ window_geom({W,H}, {_, Max, _Unit, MaxUnit},
 
 
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
-draw_win(DC, #win{no_samples=Samples, geom=#{scale:={WS,HS}}, graphs=Graphs, max={_,Max,_,_}}=Win,
+draw_win(DC, #win{name=Name, no_samples=Samples, geom=#{scale:={WS,HS}},
+                  graphs=Graphs, max={_,Max,_,_}, info=Info}=Win,
 	 #ti{tick=Tick, fetch=FetchFreq, secs=Secs, disp=DispFreq}=Ti,
-	 Paint=#paint{pens=Pens}) when Samples >= 2, Graphs =/= [] ->
+	 Paint=#paint{pens=Pens, dot_pens=Dots}) when Samples >= 2, Graphs =/= [] ->
     %% Draw graphs
     {X0,Y0,DrawBs} = draw_borders(DC, Ti, Win, Paint),
     Offset = Tick / DispFreq,
@@ -473,18 +511,27 @@ draw_win(DC, #win{no_samples=Samples, geom=#{scale:={WS,HS}}, graphs=Graphs, max
 	   end,
     Start = X0 + (max(Secs*FetchFreq+Full-Samples, 0) - Offset)*WS,
     Last = Secs*FetchFreq*WS+X0,
+    Dirty = case {Name, Info} of
+                {runq, {_, DCpu}} -> DCpu;
+                _ -> 0
+            end,
+    NoGraphs = length(Graphs),
+    NoCpu = NoGraphs - Dirty,
     Draw = fun(Lines0, N) ->
-		   setPen(DC, element(1+ ((N-1) rem tuple_size(Pens)), Pens)),
+                   case Dirty > 0 andalso N > NoCpu of
+                       true  -> setPen(DC, element(1+ ((N-NoCpu-1) rem tuple_size(Dots)), Dots));
+                       false -> setPen(DC, element(1+ ((N-1) rem tuple_size(Pens)), Pens))
+                   end,
 		   Order = lists:reverse(Lines0),
 		   [{_,Y}|Lines] = translate(Order, {Start, Y0}, 0, WS, {X0,Max*HS,Last}, []),
 		   strokeLines(DC, [{Last,Y}|Lines]),
 		   N-1
 	   end,
-    lists:foldl(Draw, length(Graphs), Graphs),
+    lists:foldl(Draw, NoGraphs, Graphs),
     DrawBs(),
     ok;
 
-draw_win(DC, #win{no_samples=Samples} = Win,Ti, #paint{small=Small}=Paint) ->
+draw_win(DC, #win{no_samples=Samples} = Win,Ti, #paint{fg=Fg, small=Small}=Paint) ->
     %% Draw Error Msg
     try draw_borders(DC, Ti, Win, Paint) of
 	{X0,_Y0,DrawBs} ->
@@ -492,7 +539,7 @@ draw_win(DC, #win{no_samples=Samples} = Win,Ti, #paint{small=Small}=Paint) ->
 		       true  -> "Waiting for data";
 		       false -> "Information not available"
 		   end,
-	    setFont(DC, Small, {0,0,0}),
+	    setFont(DC, Small, Fg),
 	    {_,WW} = getSize(DC),
 	    drawText(DC, Text, X0 + 100, WW div 2),
 	    DrawBs(),
@@ -587,7 +634,7 @@ spline_tan(Y0, Y1, Y2, Y3) ->
 
 draw_borders(DC, #ti{secs=Secs, fetch=FetchFreq},
 	     #win{name=Type, geom=Geom, info=Info, max={_,_,Unit,_}},
-	     #paint{pen=Pen, pen2=Pen2, font=Font, small=Small}) ->
+	     #paint{pen=Pen, pen2=Pen2, fg=Fg, font=Font, small=Small}) ->
     #{p0:={GraphX0, GraphY0}, p1:={GraphX1,GraphY1}, scale:={ScaleW0,_},
       txsz:={TW,TH,SpaceW}, txt:={BottomTextY, MaxTextY}, strs:={Str1,Str2,Str3}} = Geom,
 
@@ -599,7 +646,7 @@ draw_borders(DC, #ti{secs=Secs, fetch=FetchFreq},
     GraphY50 = GraphY0 + (GraphY1 - GraphY0) / 2,
     GraphY75 = GraphY0 + 3*(GraphY1 - GraphY0) / 4,
 
-    setFont(DC, Small, {0,0,0}),
+    setFont(DC, Small, Fg),
     Align = fun(Str, Y) ->
 		    {StrW, _} = getTextExtent(DC, Str),
 		    drawText(DC, Str, GraphX0 - StrW - ?BW, Y)
@@ -629,11 +676,11 @@ draw_borders(DC, #ti{secs=Secs, fetch=FetchFreq},
     strokeLine(DC, GraphX0-3, GraphY50, GraphX1, GraphY50),
     strokeLine(DC, GraphX0-3, GraphY75, GraphX1, GraphY75),
 
-    setFont(DC, Font, {0,0,0}),
+    setFont(DC, Font, Fg),
 
     Text = fun(X,Y, Str, PenId) ->
 		   if PenId == 0 ->
-			   setFont(DC, Font, {0,0,0});
+			   setFont(DC, Font, Fg);
 		      PenId > 0 ->
 			   Id = 1 + ((PenId-1) rem tuple_size(colors())),
 			   setFont(DC, Font, element(Id, colors()))
@@ -645,11 +692,17 @@ draw_borders(DC, #ti{secs=Secs, fetch=FetchFreq},
 
     case Type of
 	runq ->
+            {TextInfo, DirtyCpus} = Info,
 	    drawText(DC, "Scheduler Utilization (%) ", TopTextX, ?BH),
 	    TN0 = Text(TopTextX, BottomTextY, "Scheduler: ", 0),
-	    lists:foldl(fun(Id, Pos0) ->
-				Text(Pos0, BottomTextY, integer_to_list(Id), Id)
-			end, TN0, Info);
+            Id = fun(Id, Pos0) ->
+                         Text(Pos0, BottomTextY, integer_to_list(Id), Id)
+                 end,
+	    TN1 = lists:foldl(Id, TN0, TextInfo),
+            TN2 = Text(TN1, BottomTextY, "Dirty cpu: ", 0),
+	    TN3 = lists:foldl(Id, TN2, lists:seq(1, DirtyCpus)),
+            _ = Text(TN3, BottomTextY, "(dotted)", 0),
+            ok;
 	memory ->
 	    drawText(DC, "Memory Usage " ++ Unit, TopTextX,?BH),
 	    lists:foldl(fun(MType, {PenId, Pos0}) ->
@@ -738,10 +791,10 @@ calc_max1(Max) ->
     end.
 
 colors() ->
-    {{240, 100, 100}, {100, 240, 100}, {100, 100, 240},
-     {220, 220, 80}, {100, 240, 240}, {240, 100, 240},
-     {100, 25, 25}, {25, 100, 25}, {25, 25, 100},
-     {120, 120, 0}, {25, 100, 100}, {100, 50, 100}
+    {{240, 100, 100}, {0, 128, 0},     {25, 45, 170},  {255, 165, 0},
+     {220, 220, 40},  {100, 240, 240},{240, 100, 240}, {160, 40,  40},
+     {100, 100, 240}, {140, 140, 0},  {25, 200, 100},  {120, 25, 240},
+     {255, 140, 163}, {25, 120, 120}, {120, 25, 120},  {110, 90, 60}
     }.
 
 %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
