@@ -90,25 +90,27 @@
 
 %%----------------------------------------------------------------
 %% Types
--export_type([public_key/0,
-              private_key/0,
-              pem_entry/0,
-	      pki_asn1_type/0,
-              asn1_type/0,
-              der_encoded/0,
-              key_params/0,
-              digest_type/0,
-              issuer_name/0,
+-export_type([asn1_type/0,
+              bad_cert_reason/0,
               cert/0,
-              combined_cert/0,
               cert_id/0,
-              oid/0,
               cert_opt/0,
               chain_opts/0,
+              combined_cert/0,
               conf_opt/0,
+              der_encoded/0,
+              digest_type/0,
+              issuer_name/0,
+              key_params/0,
+              oid/0,
+              pem_entry/0,
+              pki_asn1_type/0,
+              policy_node/0,
+              private_key/0,
+              public_key/0,
               test_config/0,
-              test_root_cert/0]).
-
+              test_root_cert/0
+             ]).
 
 -type public_key()           ::  rsa_public_key() | rsa_pss_public_key() | dsa_public_key() | ec_public_key() | ed_public_key() .
 -type private_key()          ::  rsa_private_key() | rsa_pss_private_key() | dsa_private_key() | ec_private_key() | ed_private_key() .
@@ -160,11 +162,16 @@
 -type oid()                  :: tuple().
 -type cert_id()              :: {SerialNr::integer(), issuer_name()} .
 -type issuer_name()          :: {rdnSequence,[[#'AttributeTypeAndValue'{}]]} .
--type bad_cert_reason()      :: cert_expired | invalid_issuer | invalid_signature | name_not_permitted | missing_basic_constraint | invalid_key_usage | {revoked, crl_reason()} | atom().
+-type bad_cert_reason()      :: cert_expired | invalid_issuer | invalid_signature | name_not_permitted | missing_basic_constraint | invalid_key_usage | duplicate_cert_in_path |
+                                {'policy_requirement_not_met', term()} | {'invalid_policy_mapping', term()} | {revoked, crl_reason()} | invalid_validity_dates | atom().
 
 -type combined_cert()        :: #cert{}.
 -type cert()                 :: der_cert() | otp_cert().
 -type der_cert()             :: der_encoded().
+-type policy_node() ::
+        #{valid_policy := public_key:oid(),
+          qualifier_set := [#'UserNotice'{}| {uri, string()}],
+          expected_policy_set := [public_key:oid()]}.
 -type otp_cert()             :: #'OTPCertificate'{}.
 -type public_key_info()      :: {key_oid_name(),  rsa_public_key() | #'ECPoint'{} | dss_public_key(),  public_key_params()}.
 -type key_oid_name()              :: 'rsaEncryption' | 'id-RSASSA-PSS' | 'id-ecPublicKey' | 'id-Ed25519' | 'id-Ed448' | 'id-dsa'.
@@ -1134,14 +1141,14 @@ pkix_normalize_name(Issuer) ->
 
 %%-------------------------------------------------------------------- 
 -spec pkix_path_validation(Cert, CertChain, Options) ->
-          {ok, {PublicKeyInfo, PolicyTree}} |
+          {ok, {PublicKeyInfo, ConstrainedPolicyNodes}} |
           {error, {bad_cert, Reason :: bad_cert_reason()}}
               when
       Cert :: cert() | atom(),
       CertChain :: [cert() | combined_cert()],
       Options  :: [{max_path_length, integer()} | {verify_fun, {fun(), term()}}],
       PublicKeyInfo :: public_key_info(),
-      PolicyTree :: list().
+      ConstrainedPolicyNodes :: [policy_node()].
 
 %% Description: Performs a basic path validation according to RFC 5280.
 %%--------------------------------------------------------------------
@@ -1161,7 +1168,12 @@ pkix_path_validation(#'OTPCertificate'{} = TrustedCert, CertChain, Options)
                                                                 MaxPathDefault, 
                                                                 [{verify_fun, {VerifyFun, UserState1}} | 
                                                                  proplists:delete(verify_fun, Options)]),
-            path_validation(CertChain, ValidationState)
+            case exists_duplicate_cert(CertChain) of
+                true ->
+                    {error, {bad_cert, duplicate_cert_in_path}};
+                false ->
+                    path_validation(CertChain, ValidationState)
+            end
     catch
         throw:{bad_cert, _} = Result ->
             {error, Result}
@@ -1553,6 +1565,20 @@ do_pem_entry_decode({Asn1Type,_, _} = PemEntry, Password) ->
     Der = pubkey_pem:decipher(PemEntry, Password),
     der_decode(Asn1Type, Der).
 
+%% The only way a path with duplicates could be somehow wrongly
+%% passed is if the certs are located together and also are
+%% self-signed. This is what we need to possible protect against. We
+%% only check for togetherness here as it helps with the case not
+%% otherwise caught. It can result in a different error message for
+%% cases already failing before but that is not important, the
+%% important thing is that it will be rejected.
+exists_duplicate_cert([]) ->
+    false;
+exists_duplicate_cert([Cert, Cert | _]) ->
+    true;
+exists_duplicate_cert([_ | Rest]) ->
+    exists_duplicate_cert(Rest).
+
 path_validation([], #path_validation_state{working_public_key_algorithm
 					   = Algorithm,
 					   working_public_key =
@@ -1561,7 +1587,16 @@ path_validation([], #path_validation_state{working_public_key_algorithm
 					   = PublicKeyParams,
 					   valid_policy_tree = Tree
 					  }) ->
-    {ok, {{Algorithm, PublicKey, PublicKeyParams}, Tree}};
+    ValidPolicyNodeSet0 = pubkey_policy_tree:constrained_policy_node_set(Tree),
+    CollectQualifiers = fun(#{expected_policy_set := PolicySet} = Node) ->
+                                QF = fun(Policy) ->
+                                             pubkey_policy_tree:collect_qualifiers(Tree, Policy)
+                                     end,
+                                Qualifiers = lists:flatmap(QF, PolicySet),
+                                Node#{qualifier_set => Qualifiers}
+                        end,
+    ValidPolicyNodeSet =  lists:map(CollectQualifiers, ValidPolicyNodeSet0),
+    {ok, {{Algorithm, PublicKey, PublicKeyParams}, ValidPolicyNodeSet}};
 
 path_validation([DerCert | Rest], ValidationState = #path_validation_state{
 				    max_path_length = Len}) when Len >= 0 ->
@@ -1807,7 +1842,7 @@ ec_curve_spec({namedCurve, ed25519 = Name}) ->
 ec_curve_spec({namedCurve, ed448 = Name}) ->
     Name;
 ec_curve_spec({namedCurve, Name}) when is_atom(Name) ->
-    crypto:ec_curve(Name).
+   (Name).
 
 ec_curve_type(ed25519) ->
     eddsa;
